#!/usr/bin/env python3

import logging
import os
import shutil
import tempfile
import textwrap
import unittest
from pathlib import Path
from unittest import mock

from .shared_test_code import mkdtemp

import fdroidserver.common
import fdroidserver.lint
import fdroidserver.metadata
from fdroidserver._yaml import config_dump

basedir = Path(__file__).parent


class SetUpTearDownMixin:
    """A base class with no test in it for shared setUp and tearDown."""

    def setUp(self):
        os.chdir(basedir)
        fdroidserver.common.config = None
        fdroidserver.lint.config = None
        fdroidserver.lint.CATEGORIES_KEYS = None
        self._td = mkdtemp()
        self.testdir = self._td.name

    def tearDown(self):
        self._td.cleanup()


class LintTest(SetUpTearDownMixin, unittest.TestCase):
    '''fdroidserver/lint.py'''

    def test_check_for_unsupported_metadata_files(self):
        self.assertTrue(fdroidserver.lint.check_for_unsupported_metadata_files())

        with tempfile.TemporaryDirectory() as testdir:
            testdir = Path(testdir)
            self.assertFalse(
                fdroidserver.lint.check_for_unsupported_metadata_files(testdir)
            )
            shutil.copytree(
                basedir / 'metadata',
                testdir / 'metadata',
                ignore=shutil.ignore_patterns('apk', 'dump', '*.json'),
            )
            self.assertFalse(
                fdroidserver.lint.check_for_unsupported_metadata_files(testdir)
            )
            (testdir / 'metadata/org.adaway.json').write_text('placeholder')
            self.assertTrue(
                fdroidserver.lint.check_for_unsupported_metadata_files(testdir)
            )

    def test_forbidden_html_tags(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config

        app = {
            'Name': 'Bad App',
            'Summary': 'We pwn you',
            'Description': 'This way: <style><img src="</style><img src=x onerror=alert(1)//">',
        }

        anywarns = False
        for warn in fdroidserver.lint.check_regexes(app):
            anywarns = True
            logging.debug(warn)
        self.assertTrue(anywarns)

    def test_source_urls(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config

        app = {
            'Name': 'My App',
            'Summary': 'just a placeholder',
            'Description': 'This app does all sorts of useful stuff',
        }
        good_urls = [
            'https://github.com/Matteljay/mastermindy-android',
            'https://gitlab.com/origin/master',
            'https://gitlab.com/group/subgroup/masterthing',
            'https://raw.githubusercontent.com/Seva-coder/Finder/HEAD/ChangeLog.txt',
            'https://github.com/scoutant/blokish/blob/HEAD/README.md#changelog',
            'https://git.ieval.ro/?p=fonbot.git;a=blob;f=Changes;hb=HEAD',
            'https://htmlpreview.github.io/?https://github.com/YasuakiHonda/Maxima-on-Android-AS/blob/HEAD/app/src/main/assets/About_MoA/index.html',
            '',
        ]

        anywarns = False
        for url in good_urls:
            app['SourceCode'] = url
            for warn in fdroidserver.lint.check_regexes(app):
                anywarns = True
                logging.debug(warn)
        self.assertFalse(anywarns)

        bad_urls = [
            'github.com/my/proj',
            'http://github.com/not/secure',
            'https://github.com/foo/bar.git',
            'https://gitlab.com/group/subgroup/project.git',
            'https://raw.githubusercontent.com/Seva-coder/Finder/master/ChangeLog.txt',
            'https://github.com/scoutant/blokish/blob/master/README.md#changelog',
            'http://htmlpreview.github.io/?https://github.com/my/project/blob/HEAD/index.html',
            'http://fdroid.gitlab.io/fdroid-website',
        ]
        logging.debug('bad urls:')
        for url in bad_urls:
            anywarns = False
            app['SourceCode'] = url
            for warn in fdroidserver.lint.check_regexes(app):
                anywarns = True
                logging.debug(warn)
            self.assertTrue(anywarns, url + " does not fail lint!")

    def test_check_app_field_types(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config

        app = fdroidserver.metadata.App()
        app.id = 'fake.app'
        app.Name = 'Bad App'
        app.Summary = 'We pwn you'
        app.Description = 'These are some back'

        fields = {
            'Categories': {
                'good': [
                    ['Sports & Health'],
                    ['Multimedia', 'Graphics'],
                ],
                'bad': [
                    'Science & Education',
                    'Multimedia,Graphics',
                ],
            },
            'WebSite': {
                'good': [
                    'https://homepage.com',
                ],
                'bad': [
                    [],
                    [
                        'nope',
                    ],
                    29,
                ],
            },
        }

        for field, values in fields.items():
            for bad in values['bad']:
                anywarns = False
                app[field] = bad
                for warn in fdroidserver.lint.check_app_field_types(app):
                    anywarns = True
                    logging.debug(warn)
                self.assertTrue(anywarns)

            for good in values['good']:
                anywarns = False
                app[field] = good
                for warn in fdroidserver.lint.check_app_field_types(app):
                    anywarns = True
                    logging.debug(warn)
                self.assertFalse(anywarns)

    def test_check_vercode_operation(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config

        app = fdroidserver.metadata.App()
        app.Name = 'Bad App'
        app.Summary = 'We pwn you'
        app.Description = 'These are some back'

        good_fields = [
            '6%c',
            '%c - 1',
            '%c + 10',
            '%c*10',
            '%c*10 + 3',
            '%c*10 + 8',
            '%c + 2 ',
            '%c + 3',
            '%c + 7',
        ]
        bad_fields = [
            'open("/etc/passwd")',
            '%C + 1',
            '%%c * 123',
            '123 + %%',
            '%c % 7',
        ]

        anywarns = False
        for good in good_fields:
            app.VercodeOperation = [good]
            for warn in fdroidserver.lint.check_vercode_operation(app):
                anywarns = True
                logging.debug(warn)
            self.assertFalse(anywarns)

        for bad in bad_fields:
            anywarns = False
            app.VercodeOperation = [bad]
            for warn in fdroidserver.lint.check_vercode_operation(app):
                anywarns = True
                logging.debug(warn)
            self.assertTrue(anywarns)

    def test_check_license_tag_no_custom_pass(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config

        app = fdroidserver.metadata.App()
        app.License = "GPL-3.0-or-later"

        anywarns = False
        for warn in fdroidserver.lint.check_license_tag(app):
            anywarns = True
            logging.debug(warn)
        self.assertFalse(anywarns)

    def test_check_license_tag_no_custom_fail(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config

        app = fdroidserver.metadata.App()
        app.License = "Adobe-2006"

        anywarns = False
        for warn in fdroidserver.lint.check_license_tag(app):
            anywarns = True
            logging.debug(warn)
        self.assertTrue(anywarns)

    def test_check_license_tag_with_custom_pass(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config
        config['lint_licenses'] = ['fancy-license', 'GPL-3.0-or-later']

        app = fdroidserver.metadata.App()
        app.License = "fancy-license"

        anywarns = False
        for warn in fdroidserver.lint.check_license_tag(app):
            anywarns = True
            logging.debug(warn)
        self.assertFalse(anywarns)

    def test_check_license_tag_with_custom_fail(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config
        config['lint_licenses'] = ['fancy-license', 'GPL-3.0-or-later']

        app = fdroidserver.metadata.App()
        app.License = "Apache-2.0"

        anywarns = False
        for warn in fdroidserver.lint.check_license_tag(app):
            anywarns = True
            logging.debug(warn)
        self.assertTrue(anywarns)

    def test_check_license_tag_with_custom_empty(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config
        config['lint_licenses'] = []

        app = fdroidserver.metadata.App()
        app.License = "Apache-2.0"

        anywarns = False
        for warn in fdroidserver.lint.check_license_tag(app):
            anywarns = True
            logging.debug(warn)
        self.assertTrue(anywarns)

    def test_check_license_tag_disabled(self):
        config = dict()
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.common.config = config
        fdroidserver.lint.config = config
        config['lint_licenses'] = None

        app = fdroidserver.metadata.App()
        app.License = "Apache-2.0"

        anywarns = False
        for warn in fdroidserver.lint.check_license_tag(app):
            anywarns = True
            logging.debug(warn)
        self.assertFalse(anywarns)

    def test_check_categories_in_config(self):
        fdroidserver.lint.config = {
            fdroidserver.common.CATEGORIES_CONFIG_NAME: ['InConfig']
        }
        fdroidserver.lint.load_categories_config()
        app = fdroidserver.metadata.App({'Categories': ['InConfig']})
        self.assertEqual(0, len(list(fdroidserver.lint.check_categories(app))))

    def test_check_categories_not_in_config(self):
        fdroidserver.lint.config = dict()
        fdroidserver.lint.load_categories_config()
        app = fdroidserver.metadata.App({'Categories': ['NotInConfig']})
        self.assertEqual(1, len(list(fdroidserver.lint.check_categories(app))))

    def test_check_categories_empty_is_error(self):
        fdroidserver.lint.config = {fdroidserver.common.CATEGORIES_CONFIG_NAME: []}
        fdroidserver.lint.load_categories_config()
        app = fdroidserver.metadata.App({'Categories': ['something']})
        self.assertEqual(1, len(list(fdroidserver.lint.check_categories(app))))

    def test_check_categories_old_hardcoded_not_defined(self):
        fdroidserver.lint.config = {
            fdroidserver.common.CATEGORIES_CONFIG_NAME: ['foo', 'bar']
        }
        fdroidserver.lint.load_categories_config()
        app = fdroidserver.metadata.App({'Categories': ['Writing']})
        self.assertEqual(1, len(list(fdroidserver.lint.check_categories(app))))

    def test_check_categories_from_config_yml(self):
        """In config.yml, categories is a list."""
        os.chdir(self.testdir)
        fdroidserver.common.write_config_file('categories: [foo, bar]\n')
        fdroidserver.lint.config = fdroidserver.common.read_config()
        fdroidserver.lint.load_categories_config()
        self.assertEqual(fdroidserver.lint.CATEGORIES_KEYS, ['foo', 'bar'])
        app = fdroidserver.metadata.App({'Categories': ['bar']})
        self.assertEqual(0, len(list(fdroidserver.lint.check_categories(app))))

    def test_check_categories_from_config_categories_yml(self):
        """In config/categories.yml, categories is a localized STRINGMAP dict."""
        os.chdir(self.testdir)
        os.mkdir('config')
        Path('config/categories.yml').write_text('{foo: {name: foo}, bar: {name: bar}}')
        fdroidserver.lint.config = fdroidserver.common.read_config()
        fdroidserver.lint.load_categories_config()
        self.assertEqual(fdroidserver.lint.CATEGORIES_KEYS, ['foo', 'bar'])
        app = fdroidserver.metadata.App({'Categories': ['bar']})
        self.assertEqual(0, len(list(fdroidserver.lint.check_categories(app))))

    def test_lint_config_basic_mirrors_yml(self):
        os.chdir(self.testdir)
        with Path('mirrors.yml').open('w') as fp:
            config_dump([{'url': 'https://example.com/fdroid/repo'}], fp)
        self.assertTrue(fdroidserver.lint.lint_config('mirrors.yml'))

    def test_lint_config_mirrors_yml_kenya_countryCode(self):
        os.chdir(self.testdir)
        with Path('mirrors.yml').open('w') as fp:
            config_dump(
                [{'url': 'https://foo.com/fdroid/repo', 'countryCode': 'KE'}], fp
            )
        self.assertTrue(fdroidserver.lint.lint_config('mirrors.yml'))

    def test_lint_config_mirrors_yml_invalid_countryCode(self):
        """WV is "indeterminately reserved" so it should never be used."""
        os.chdir(self.testdir)
        with Path('mirrors.yml').open('w') as fp:
            config_dump(
                [{'url': 'https://foo.com/fdroid/repo', 'countryCode': 'WV'}], fp
            )
        self.assertFalse(fdroidserver.lint.lint_config('mirrors.yml'))

    def test_lint_config_mirrors_yml_alpha3_countryCode(self):
        """Only ISO 3166-1 alpha 2 are supported"""
        os.chdir(self.testdir)
        with Path('mirrors.yml').open('w') as fp:
            config_dump(
                [{'url': 'https://de.com/fdroid/repo', 'countryCode': 'DEU'}], fp
            )
        self.assertFalse(fdroidserver.lint.lint_config('mirrors.yml'))

    def test_lint_config_mirrors_yml_one_invalid_countryCode(self):
        """WV is "indeterminately reserved" so it should never be used."""
        os.chdir(self.testdir)
        with Path('mirrors.yml').open('w') as fp:
            config_dump(
                [
                    {'url': 'https://bar.com/fdroid/repo', 'countryCode': 'BA'},
                    {'url': 'https://foo.com/fdroid/repo', 'countryCode': 'FO'},
                    {'url': 'https://wv.com/fdroid/repo', 'countryCode': 'WV'},
                ],
                fp,
            )
        self.assertFalse(fdroidserver.lint.lint_config('mirrors.yml'))

    def test_lint_config_bad_mirrors_yml_dict(self):
        os.chdir(self.testdir)
        Path('mirrors.yml').write_text('baz: [foo, bar]\n')
        with self.assertRaises(TypeError):
            fdroidserver.lint.lint_config('mirrors.yml')

    def test_lint_config_bad_mirrors_yml_float(self):
        os.chdir(self.testdir)
        Path('mirrors.yml').write_text('1.0\n')
        with self.assertRaises(TypeError):
            fdroidserver.lint.lint_config('mirrors.yml')

    def test_lint_config_bad_mirrors_yml_int(self):
        os.chdir(self.testdir)
        Path('mirrors.yml').write_text('1\n')
        with self.assertRaises(TypeError):
            fdroidserver.lint.lint_config('mirrors.yml')

    def test_lint_config_bad_mirrors_yml_str(self):
        os.chdir(self.testdir)
        Path('mirrors.yml').write_text('foo\n')
        with self.assertRaises(TypeError):
            fdroidserver.lint.lint_config('mirrors.yml')

    def test_lint_invalid_config_keys(self):
        os.chdir(self.testdir)
        os.mkdir('config')
        config_yml = fdroidserver.common.CONFIG_FILE
        with open(f'config/{config_yml}', 'w', encoding='utf-8') as fp:
            fp.write('repo:\n  invalid_key: test\n')
        self.assertFalse(fdroidserver.lint.lint_config(f'config/{config_yml}'))

    def test_lint_invalid_localized_config_keys(self):
        os.chdir(self.testdir)
        Path('config/en').mkdir(parents=True)
        Path('config/en/antiFeatures.yml').write_text('NonFreeNet:\n  icon: test.png\n')
        self.assertFalse(fdroidserver.lint.lint_config('config/en/antiFeatures.yml'))

    def test_check_certificate_pinned_binaries_empty(self):
        fdroidserver.common.config = {}
        app = fdroidserver.metadata.App()
        app.AllowedAPKSigningKeys = [
            'a40da80a59d170caa950cf15c18c454d47a39b26989d8b640ecd745ba71bf5dc'
        ]
        self.assertEqual(
            [],
            list(fdroidserver.lint.check_certificate_pinned_binaries(app)),
            "when the config is empty, any signing key should be allowed",
        )

    def test_lint_known_debug_keys_no_match(self):
        fdroidserver.common.config = {
            "apk_signing_key_block_list": "a40da80a59d170caa950cf15c18c454d47a39b26989d8b640ecd745ba71bf5dc"
        }
        app = fdroidserver.metadata.App()
        app.AllowedAPKSigningKeys = [
            '2fd4fd5f54babba4bcb21237809bb653361d0d2583c80964ec89b28a26e9539e'
        ]
        self.assertEqual(
            [],
            list(fdroidserver.lint.check_certificate_pinned_binaries(app)),
            "A signing key that does not match one in the config should be allowed",
        )

    def test_lint_known_debug_keys(self):
        fdroidserver.common.config = {
            'apk_signing_key_block_list': 'a40da80a59d170caa950cf15c18c454d47a39b26989d8b640ecd745ba71bf5dc'
        }
        app = fdroidserver.metadata.App()
        app.AllowedAPKSigningKeys = [
            'a40da80a59d170caa950cf15c18c454d47a39b26989d8b640ecd745ba71bf5dc'
        ]
        for warn in fdroidserver.lint.check_certificate_pinned_binaries(app):
            anywarns = True
            logging.debug(warn)
        self.assertTrue(anywarns)


class LintAntiFeaturesTest(unittest.TestCase):
    def setUp(self):
        os.chdir(basedir)
        fdroidserver.common.config = dict()
        fdroidserver.lint.ANTIFEATURES_KEYS = None
        fdroidserver.lint.load_antiFeatures_config()

    def test_check_antiFeatures_empty(self):
        app = fdroidserver.metadata.App()
        self.assertEqual([], list(fdroidserver.lint.check_antiFeatures(app)))

    def test_check_antiFeatures_empty_AntiFeatures(self):
        app = fdroidserver.metadata.App()
        app['AntiFeatures'] = []
        self.assertEqual([], list(fdroidserver.lint.check_antiFeatures(app)))

    def test_check_antiFeatures(self):
        app = fdroidserver.metadata.App()
        app['AntiFeatures'] = ['Ads', 'UpstreamNonFree']
        self.assertEqual([], list(fdroidserver.lint.check_antiFeatures(app)))

    def test_check_antiFeatures_fails_one(self):
        app = fdroidserver.metadata.App()
        app['AntiFeatures'] = ['Ad']
        self.assertEqual(1, len(list(fdroidserver.lint.check_antiFeatures(app))))

    def test_check_antiFeatures_fails_many(self):
        app = fdroidserver.metadata.App()
        app['AntiFeatures'] = ['Adss', 'Tracker', 'NoSourceSince', 'FAKE', 'NonFree']
        self.assertEqual(4, len(list(fdroidserver.lint.check_antiFeatures(app))))

    def test_check_antiFeatures_build_empty(self):
        app = fdroidserver.metadata.App()
        app['Builds'] = [{'antifeatures': []}]
        self.assertEqual([], list(fdroidserver.lint.check_antiFeatures(app)))

    def test_check_antiFeatures_build(self):
        app = fdroidserver.metadata.App()
        app['Builds'] = [{'antifeatures': ['Tracking']}]
        self.assertEqual(0, len(list(fdroidserver.lint.check_antiFeatures(app))))

    def test_check_antiFeatures_build_fail(self):
        app = fdroidserver.metadata.App()
        app['Builds'] = [{'antifeatures': ['Ads', 'Tracker']}]
        self.assertEqual(1, len(list(fdroidserver.lint.check_antiFeatures(app))))


class ConfigYmlTest(LintTest):
    """Test data formats used in config.yml.

    lint.py uses print() and not logging so hacks are used to control
    the output when running in the test runner.

    """

    def setUp(self):
        super().setUp()
        self.config_yml = Path(self.testdir) / fdroidserver.common.CONFIG_FILE

    def test_config_yml_int(self):
        self.config_yml.write_text('repo_maxage: 1\n')
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    @mock.patch('builtins.print', mock.Mock())  # hide error message
    def test_config_yml_int_bad(self):
        self.config_yml.write_text('repo_maxage: "1"\n')
        self.assertFalse(fdroidserver.lint.lint_config(self.config_yml))

    def test_config_yml_str(self):
        self.config_yml.write_text('sdk_path: /opt/android-sdk\n')
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    def test_config_yml_str_list(self):
        self.config_yml.write_text('serverwebroot: [server1, server2]\n')
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    def test_config_yml_str_list_of_dicts(self):
        self.config_yml.write_text(
            textwrap.dedent(
                """\
                serverwebroot:
                  - url: 'me@b.az:/srv/fdroid'
                    index_only: true
                """
            )
        )
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    def test_config_yml_str_list_of_dicts_env(self):
        """serverwebroot can be str, list of str, or list of dicts."""
        self.config_yml.write_text('serverwebroot: {env: ANDROID_HOME}\n')
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    def test_config_yml_str_env(self):
        self.config_yml.write_text('sdk_path: {env: ANDROID_HOME}\n')
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    @mock.patch('builtins.print', mock.Mock())  # hide error message
    def test_config_yml_str_bad(self):
        self.config_yml.write_text('sdk_path: 1.0\n')
        self.assertFalse(fdroidserver.lint.lint_config(self.config_yml))

    def test_config_yml_bool(self):
        self.config_yml.write_text("deploy_process_logs: true\n")
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    @mock.patch('builtins.print', mock.Mock())  # hide error message
    def test_config_yml_bool_bad(self):
        self.config_yml.write_text('deploy_process_logs: 2342fe23\n')
        self.assertFalse(fdroidserver.lint.lint_config(self.config_yml))

    def test_config_yml_dict(self):
        self.config_yml.write_text("keyaliases: {com.example: '@com.foo'}\n")
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    @mock.patch('builtins.print', mock.Mock())  # hide error message
    def test_config_yml_dict_bad(self):
        self.config_yml.write_text('keyaliases: 2342fe23\n')
        self.assertFalse(fdroidserver.lint.lint_config(self.config_yml))

    @mock.patch('builtins.print', mock.Mock())  # hide error message
    def test_config_yml_bad_key_name(self):
        self.config_yml.write_text('keyalias: 2342fe23\n')
        self.assertFalse(fdroidserver.lint.lint_config(self.config_yml))

    @mock.patch('builtins.print', mock.Mock())  # hide error message
    def test_config_yml_bad_value_for_all_keys(self):
        """Check all config keys with a bad value."""
        for key in fdroidserver.lint.check_config_keys:
            if key in fdroidserver.lint.bool_keys:
                value = 'foobar'
            else:
                value = 'false'
            self.config_yml.write_text(f'{key}: {value}\n')
            self.assertFalse(
                fdroidserver.lint.lint_config(self.config_yml),
                f'{key} should fail on value of "{value}"',
            )

    def test_config_yml_keyaliases(self):
        self.config_yml.write_text(
            textwrap.dedent(
                """\
                keyaliases:
                  com.example: myalias
                  com.foo: '@com.example'
                """
            )
        )
        self.assertTrue(fdroidserver.lint.lint_config(self.config_yml))

    @mock.patch('builtins.print', mock.Mock())  # hide error message
    def test_config_yml_keyaliases_bad_str(self):
        """The keyaliases: value is a dict not a str."""
        self.config_yml.write_text("keyaliases: '@com.example'\n")
        self.assertFalse(fdroidserver.lint.lint_config(self.config_yml))

    @mock.patch('builtins.print', mock.Mock())  # hide error message
    def test_config_yml_keyaliases_bad_list(self):
        """The keyaliases: value is a dict not a list."""
        self.config_yml.write_text(
            textwrap.dedent(
                """\
                keyaliases:
                  - com.example: myalias
                """
            )
        )
        self.assertFalse(fdroidserver.lint.lint_config(self.config_yml))



#!/usr/bin/env python3
"""Extract application metadata from a source repository."""
#
# import_subcommand.py - part of the FDroid server tools
# Copyright (C) 2010-13, Ciaran Gultnieks, ciaran@ciarang.com
# Copyright (C) 2013-2014 Daniel Martí <mvdan@mvdan.cc>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import re
import shutil
import stat
import sys
import urllib
from argparse import ArgumentParser
from pathlib import Path
from typing import Optional

import git
import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

from . import _, common, metadata
from .exception import FDroidException

config = None

SETTINGS_GRADLE_REGEX = re.compile(r'settings\.gradle(?:\.kts)?')
GRADLE_SUBPROJECT_REGEX = re.compile(r'''['"]:?([^'"]+)['"]''')
APPLICATION_ID_REGEX = re.compile(r'''\s*applicationId\s=?\s?['"].*['"]''')


def get_all_gradle_and_manifests(build_dir):
    paths = []
    for root, dirs, files in os.walk(build_dir):
        for f in sorted(files):
            if f == 'AndroidManifest.xml' or f.endswith(('.gradle', '.gradle.kts')):
                full = Path(root) / f
                paths.append(full)
    return paths


def get_gradle_subdir(build_dir, paths):
    """Get the subdir where the gradle build is based."""
    first_gradle_dir = None
    for path in paths:
        if not first_gradle_dir:
            first_gradle_dir = path.parent.relative_to(build_dir)
        if path.exists() and SETTINGS_GRADLE_REGEX.match(path.name):
            for m in GRADLE_SUBPROJECT_REGEX.finditer(path.read_text(encoding='utf-8')):
                for f in (path.parent / m.group(1)).glob('build.gradle*'):
                    with f.open(encoding='utf-8') as fp:
                        for line in fp:
                            if common.ANDROID_PLUGIN_REGEX.match(
                                line
                            ) or APPLICATION_ID_REGEX.match(line):
                                return f.parent.relative_to(build_dir)
    if first_gradle_dir and first_gradle_dir != Path('.'):
        return first_gradle_dir


def handle_retree_error_on_windows(function, path, excinfo):
    """Python can't remove a readonly file on Windows so chmod first."""
    if function in (os.unlink, os.rmdir, os.remove) and excinfo[0] == PermissionError:
        os.chmod(path, stat.S_IWRITE)
        function(path)


def clone_to_tmp_dir(app: metadata.App, rev=None) -> Path:
    """Clone the source repository of an app to a temporary directory for further processing.

    Parameters
    ----------
    app
        The App instance to clone the source of.

    Returns
    -------
    tmp_dir
        The (temporary) directory the apps source has been cloned into.

    """
    tmp_dir = Path('tmp')
    tmp_dir.mkdir(exist_ok=True)

    tmp_dir = tmp_dir / 'importer'

    if tmp_dir.exists():
        shutil.rmtree(str(tmp_dir), onerror=handle_retree_error_on_windows)
    vcs = common.getvcs(app.RepoType, app.Repo, tmp_dir)
    vcs.gotorevision(rev)

    return tmp_dir


def getrepofrompage(url: str) -> tuple[Optional[str], str]:
    """Get the repo type and address from the given web page.

    The page is scanned in a rather naive manner for 'git clone xxxx',
    'hg clone xxxx', etc, and when one of these is found it's assumed
    that's the information we want.  Returns repotype, address, or
    None, reason

    Parameters
    ----------
    url
        The url to look for repository information at.

    Returns
    -------
    repotype_or_none
        The found repository type or None if an error occured.
    address_or_reason
        The address to the found repository or the reason if an error occured.

    """
    if not url.startswith('http'):
        return (None, _('{url} does not start with "http"!'.format(url=url)))
    req = urllib.request.urlopen(url)  # nosec B310 non-http URLs are filtered out
    if req.getcode() != 200:
        return (None, 'Unable to get ' + url + ' - return code ' + str(req.getcode()))
    page = req.read().decode(req.headers.get_content_charset())

    # Works for BitBucket
    m = re.search('data-fetch-url="(.*)"', page)
    if m is not None:
        repo = m.group(1)

        if repo.endswith('.git'):
            return ('git', repo)

        return ('hg', repo)

    # Works for BitBucket (obsolete)
    index = page.find('hg clone')
    if index != -1:
        repotype = 'hg'
        repo = page[index + 9 :]
        index = repo.find('<')
        if index == -1:
            return (None, _("Error while getting repo address"))
        repo = repo[:index]
        repo = repo.split('"')[0]
        return (repotype, repo)

    # Works for BitBucket (obsolete)
    index = page.find('git clone')
    if index != -1:
        repotype = 'git'
        repo = page[index + 10 :]
        index = repo.find('<')
        if index == -1:
            return (None, _("Error while getting repo address"))
        repo = repo[:index]
        repo = repo.split('"')[0]
        return (repotype, repo)

    return (None, _("No information found.") + page)


def get_app_from_url(url: str) -> metadata.App:
    """Guess basic app metadata from the URL.

    The URL must include a network hostname, unless it is an lp:,
    file:, or git/ssh URL.  This throws ValueError on bad URLs to
    match urlparse().

    Parameters
    ----------
    url
        The URL to look to look for app metadata at.

    Returns
    -------
    app
        App instance with the found metadata.

    Raises
    ------
    :exc:`~fdroidserver.exception.FDroidException`
        If the VCS type could not be determined.
    :exc:`ValueError`
        If the URL is invalid.

    """
    parsed = urllib.parse.urlparse(url)
    invalid_url = False
    if not parsed.scheme or not parsed.path:
        invalid_url = True

    app = metadata.App()
    app.Repo = url
    if url.startswith('git://') or url.startswith('git@'):
        app.RepoType = 'git'
    elif parsed.netloc == 'github.com':
        app.RepoType = 'git'
        app.SourceCode = url
        app.IssueTracker = url + '/issues'
    elif parsed.netloc in ('gitlab.com', 'framagit.org'):
        # git can be fussy with gitlab URLs unless they end in .git
        if url.endswith('.git'):
            url = url[:-4]
        app.Repo = url + '.git'
        app.RepoType = 'git'
        app.SourceCode = url
        app.IssueTracker = url + '/issues'
    elif parsed.netloc == 'notabug.org':
        if url.endswith('.git'):
            url = url[:-4]
        app.Repo = url + '.git'
        app.RepoType = 'git'
        app.SourceCode = url
        app.IssueTracker = url + '/issues'
    elif parsed.netloc == 'bitbucket.org':
        if url.endswith('/'):
            url = url[:-1]
        app.SourceCode = url + '/src'
        app.IssueTracker = url + '/issues'
        # Figure out the repo type and adddress...
        app.RepoType, app.Repo = getrepofrompage(url)
    elif parsed.netloc == 'codeberg.org':
        app.RepoType = 'git'
        app.SourceCode = url
        app.IssueTracker = url + '/issues'
    elif url.startswith('https://') and url.endswith('.git'):
        app.RepoType = 'git'

    if not parsed.netloc and parsed.scheme in ('git', 'http', 'https', 'ssh'):
        invalid_url = True

    if invalid_url:
        raise ValueError(_('"{url}" is not a valid URL!'.format(url=url)))

    if not app.RepoType:
        raise FDroidException("Unable to determine vcs type. " + app.Repo)

    return app


def main():
    """Extract app metadata and write it to a file.

    The behaviour of this function is influenced by the configuration file as
    well as command line parameters.

    Raises
    ------
    :exc:`~fdroidserver.exception.FDroidException`
        If the repository already has local metadata, no URL is specified and
        the current directory is not a Git repository, no application ID could
        be found, no Gradle project could be found or there is already metadata
        for the found application ID.

    """
    global config

    # Parse command line...
    parser = ArgumentParser()
    common.setup_global_opts(parser)
    parser.add_argument("-u", "--url", help=_("Project URL to import from."))
    parser.add_argument(
        "-s",
        "--subdir",
        help=_("Path to main Android project subdirectory, if not in root."),
    )
    parser.add_argument(
        "-c",
        "--categories",
        help=_("Comma separated list of categories."),
    )
    parser.add_argument("-l", "--license", help=_("Overall license of the project."))
    parser.add_argument(
        "--omit-disable",
        action="store_true",
        help=_("Do not add 'disable:' to the generated build entries"),
    )
    parser.add_argument(
        "--rev",
        help=_(
            "Allows a different revision (or git branch) to be specified for the initial import"
        ),
    )
    metadata.add_metadata_arguments(parser)
    options = common.parse_args(parser)
    metadata.warnings_action = options.W

    config = common.read_config()

    apps = metadata.read_metadata()
    app = None

    tmp_importer_dir = None

    local_metadata_files = common.get_local_metadata_files()
    if local_metadata_files:
        raise FDroidException(
            _("This repo already has local metadata: %s") % local_metadata_files[0]
        )

    build = metadata.Build()
    app = metadata.App()
    if options.url is None and Path('.git').is_dir():
        app.RepoType = 'git'
        tmp_importer_dir = Path.cwd()
        git_repo = git.Repo(tmp_importer_dir)
        for remote in git.Remote.iter_items(git_repo):
            if remote.name == 'origin':
                url = git_repo.remotes.origin.url
                app = get_app_from_url(url)
                break
        write_local_file = True
    elif options.url:
        app = get_app_from_url(options.url)
        tmp_importer_dir = clone_to_tmp_dir(app, options.rev)
        git_repo = git.Repo(tmp_importer_dir)

        if not options.omit_disable:
            build.disable = (
                'Generated by `fdroid import` - check version fields and commitid'
            )
        write_local_file = False
    else:
        raise FDroidException("Specify project url.")

    app.AutoUpdateMode = 'Version'
    app.UpdateCheckMode = 'Tags'
    build.commit = common.get_head_commit_id(tmp_importer_dir)

    # Extract some information...
    paths = get_all_gradle_and_manifests(tmp_importer_dir)
    gradle_subdir = get_gradle_subdir(tmp_importer_dir, paths)
    if paths:
        versionName, versionCode, appid = common.parse_androidmanifests(paths, app)
        if not appid:
            raise FDroidException(_("Couldn't find Application ID"))
        if not versionName:
            logging.warning(_('Could not find latest version name'))
        if not versionCode:
            logging.warning(_('Could not find latest version code'))
    else:
        raise FDroidException(_("No gradle project could be found. Specify --subdir?"))

    # Make sure it's actually new...
    if appid in apps:
        raise FDroidException(_('Package "{appid}" already exists').format(appid=appid))

    # Create a build line...
    build.versionName = versionName or 'Unknown'
    app.CurrentVersion = build.versionName
    build.versionCode = versionCode or 0
    app.CurrentVersionCode = build.versionCode
    if options.subdir:
        build.subdir = options.subdir
    elif gradle_subdir:
        build.subdir = gradle_subdir.as_posix()
    # subdir might be None
    subdir = Path(tmp_importer_dir / build.subdir) if build.subdir else tmp_importer_dir

    if options.license:
        app.License = options.license
    if options.categories:
        app.Categories = options.categories.split(',')
    if (subdir / 'jni').exists():
        build.buildjni = ['yes']
    if (subdir / 'build.gradle').exists() or (subdir / 'build.gradle.kts').exists():
        build.gradle = ['yes']

    app.AutoName = common.fetch_real_name(subdir, build.gradle)

    package_json = tmp_importer_dir / 'package.json'  # react-native
    pubspec_yaml = tmp_importer_dir / 'pubspec.yaml'  # flutter
    if package_json.exists():
        build.sudo = [
            'sysctl fs.inotify.max_user_watches=524288 || true',
            'apt-get update',
            'apt-get install -y npm',
        ]
        build.init = ['npm install --build-from-source']
        with package_json.open() as fp:
            data = json.load(fp)
        app.AutoName = app.AutoName or data.get('name')
        app.License = data.get('license', app.License)
        app.Description = data.get('description', app.Description)
        app.WebSite = data.get('homepage', app.WebSite)
        app_json = tmp_importer_dir / 'app.json'
        build.scanignore = ['android/build.gradle']
        build.scandelete = ['node_modules']
        if app_json.exists():
            with app_json.open() as fp:
                data = json.load(fp)
            app.AutoName = app.AutoName or data.get('name')
    if pubspec_yaml.exists():
        with pubspec_yaml.open() as fp:
            data = yaml.load(fp, Loader=SafeLoader)
        app.AutoName = app.AutoName or data.get('name')
        app.License = data.get('license', app.License)
        app.Description = data.get('description', app.Description)
        app.UpdateCheckData = 'pubspec.yaml|version:\\s.+\\+(\\d+)|.|version:\\s(.+)\\+'
        build.srclibs = ['flutter@stable']
        build.output = 'build/app/outputs/flutter-apk/app-release.apk'
        build.subdir = None
        build.gradle = None
        build.prebuild = [
            'export PUB_CACHE=$(pwd)/.pub-cache',
            '$$flutter$$/bin/flutter config --no-analytics',
            '$$flutter$$/bin/flutter packages pub get',
        ]
        build.scandelete = [
            '.pub-cache',
        ]
        build.build = [
            'export PUB_CACHE=$(pwd)/.pub-cache',
            '$$flutter$$/bin/flutter build apk',
        ]

    git_modules = tmp_importer_dir / '.gitmodules'
    if git_modules.exists():
        build.submodules = True

    metadata.post_parse_yaml_metadata(app)

    app['Builds'].append(build)

    if write_local_file:
        metadata.write_metadata(Path('.fdroid.yml'), app)
    else:
        # Keep the repo directory to save bandwidth...
        Path('build').mkdir(exist_ok=True)
        build_dir = Path('build') / appid
        if build_dir.exists():
            logging.warning(
                _('{path} already exists, ignoring import results!').format(
                    path=build_dir
                )
            )
            sys.exit(1)
        elif tmp_importer_dir:
            # For Windows: Close the repo or a git.exe instance holds handles to repo
            try:
                git_repo.close()
            except AttributeError:  # Debian/stretch's version does not have close()
                pass
            shutil.move(tmp_importer_dir, build_dir)
        Path('build/.fdroidvcs-' + appid).write_text(app.RepoType + ' ' + app.Repo)

        metadatapath = Path('metadata') / (appid + '.yml')
        metadata.write_metadata(metadatapath, app)
        logging.info("Wrote " + str(metadatapath))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import configparser
import os
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest import mock

import git

import fdroidserver
from .shared_test_code import TmpCwd, mkdtemp, VerboseFalseOptions

basedir = Path(__file__).parent


class DeployTest(unittest.TestCase):
    '''fdroidserver/deploy.py'''

    def setUp(self):
        os.chdir(basedir)
        self._td = mkdtemp()
        self.testdir = self._td.name

        fdroidserver.common.options = mock.Mock()
        fdroidserver.deploy.config = {}
        fdroidserver.deploy.USER_RCLONE_CONF = False

    def tearDown(self):
        self._td.cleanup()

    def test_update_serverwebroots_bad_None(self):
        with self.assertRaises(TypeError):
            fdroidserver.deploy.update_serverwebroots(None, 'repo')

    def test_update_serverwebroots_bad_int(self):
        with self.assertRaises(TypeError):
            fdroidserver.deploy.update_serverwebroots(9, 'repo')

    def test_update_serverwebroots_bad_float(self):
        with self.assertRaises(TypeError):
            fdroidserver.deploy.update_serverwebroots(1.0, 'repo')

    def test_update_serverwebroots(self):
        """rsync works with file paths, so this test uses paths for the URLs"""
        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir()
        fake_apk = repo / 'fake.apk'
        with fake_apk.open('w') as fp:
            fp.write('not an APK, but has the right filename')
        url0 = Path('url0/fdroid')
        url0.mkdir(parents=True)
        url1 = Path('url1/fdroid')
        url1.mkdir(parents=True)

        # setup parameters for this test run
        fdroidserver.common.options.identity_file = None
        fdroidserver.deploy.config['make_current_version_link'] = False

        dest_apk0 = url0 / fake_apk
        dest_apk1 = url1 / fake_apk
        self.assertFalse(dest_apk0.is_file())
        self.assertFalse(dest_apk1.is_file())
        fdroidserver.deploy.update_serverwebroots(
            [
                {'url': str(url0)},
                {'url': str(url1)},
            ],
            str(repo),
        )
        self.assertTrue(dest_apk0.is_file())
        self.assertTrue(dest_apk1.is_file())

    def test_update_serverwebroots_url_does_not_end_with_fdroid(self):
        with self.assertRaises(SystemExit):
            fdroidserver.deploy.update_serverwebroots([{'url': 'url'}], 'repo')

    def test_update_serverwebroots_bad_ssh_url(self):
        with self.assertRaises(SystemExit):
            fdroidserver.deploy.update_serverwebroots(
                [{'url': 'f@b.ar::/path/to/fdroid'}], 'repo'
            )

    def test_update_serverwebroots_unsupported_ssh_url(self):
        with self.assertRaises(SystemExit):
            fdroidserver.deploy.update_serverwebroots([{'url': 'ssh://nope'}], 'repo')

    @unittest.skipUnless(shutil.which('rclone'), '/usr/bin/rclone')
    def test_update_remote_storage_with_rclone(self):
        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir(parents=True, exist_ok=True)

        fake_apk = repo / 'another_fake.apk'
        with fake_apk.open('w') as fp:
            fp.write('not an APK, but has the right filename')
        fake_index = repo / fdroidserver.common.INDEX_FILES[0]
        with fake_index.open('w') as fp:
            fp.write('not an index, but has the right filename')

        # write out rclone config for test use
        rclone_config = configparser.ConfigParser()
        rclone_config.add_section("test-local-config")
        rclone_config.set("test-local-config", "type", "local")

        rclone_config_path = Path('rclone_config_path')
        rclone_config_path.mkdir(parents=True, exist_ok=True)
        rclone_file = rclone_config_path / 'rclone.conf'
        with open(rclone_file, 'w') as configfile:
            rclone_config.write(configfile)

        # setup parameters for this test run
        fdroidserver.deploy.config['awsbucket'] = 'test_bucket_folder'
        fdroidserver.deploy.config['rclone'] = True
        fdroidserver.deploy.config['rclone_config'] = 'test-local-config'
        fdroidserver.deploy.config['path_to_custom_rclone_config'] = str(rclone_file)
        fdroidserver.common.options = VerboseFalseOptions

        # write out destination path
        destination = Path('test_bucket_folder/fdroid')
        destination.mkdir(parents=True, exist_ok=True)
        dest_apk = Path(destination) / fake_apk
        dest_index = Path(destination) / fake_index
        self.assertFalse(dest_apk.is_file())
        self.assertFalse(dest_index.is_file())
        repo_section = str(repo)
        # fdroidserver.deploy.USER_RCLONE_CONF = str(rclone_file)
        fdroidserver.deploy.update_remote_storage_with_rclone(repo_section)
        self.assertTrue(dest_apk.is_file())
        self.assertTrue(dest_index.is_file())

    @unittest.skipUnless(shutil.which('rclone'), '/usr/bin/rclone')
    def test_update_remote_storage_with_rclone_in_index_only_mode(self):
        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir(parents=True, exist_ok=True)

        fake_apk = repo / 'another_fake.apk'
        with fake_apk.open('w') as fp:
            fp.write('not an APK, but has the right filename')
        fake_index = repo / fdroidserver.common.INDEX_FILES[0]
        with fake_index.open('w') as fp:
            fp.write('not an index, but has the right filename')

        # write out rclone config for test use
        rclone_config = configparser.ConfigParser()
        rclone_config.add_section("test-local-config")
        rclone_config.set("test-local-config", "type", "local")

        rclone_config_path = Path('rclone_config_path')
        rclone_config_path.mkdir(parents=True, exist_ok=True)
        rclone_file = rclone_config_path / 'rclone.conf'
        with open(rclone_file, 'w') as configfile:
            rclone_config.write(configfile)

        # setup parameters for this test run
        fdroidserver.deploy.config['awsbucket'] = 'test_bucket_folder'
        fdroidserver.deploy.config['rclone'] = True
        fdroidserver.deploy.config['rclone_config'] = 'test-local-config'
        fdroidserver.deploy.config['path_to_custom_rclone_config'] = str(rclone_file)
        fdroidserver.common.options = VerboseFalseOptions

        # write out destination path
        destination = Path('test_bucket_folder/fdroid')
        destination.mkdir(parents=True, exist_ok=True)
        dest_apk = Path(destination) / fake_apk
        dest_index = Path(destination) / fake_index
        self.assertFalse(dest_apk.is_file())
        self.assertFalse(dest_index.is_file())
        repo_section = str(repo)
        # fdroidserver.deploy.USER_RCLONE_CONF = str(rclone_file)
        fdroidserver.deploy.update_remote_storage_with_rclone(
            repo_section, is_index_only=True
        )
        self.assertFalse(dest_apk.is_file())
        self.assertTrue(dest_index.is_file())

    @mock.patch('subprocess.call')
    @mock.patch('subprocess.check_output', lambda cmd, text: '/path/to/rclone.conf')
    def test_update_remote_storage_with_rclone_mock(self, mock_call):
        def _mock_subprocess_call(cmd):
            self.assertEqual(
                cmd,
                [
                    'rclone',
                    'sync',
                    'repo',
                    'test_local_config:test_bucket_folder/fdroid/repo',
                ],
            )
            return 0

        mock_call.side_effect = _mock_subprocess_call

        fdroidserver.deploy.config = {
            'awsbucket': 'test_bucket_folder',
            'rclone': True,
            'rclone_config': 'test_local_config',
        }
        fdroidserver.deploy.update_remote_storage_with_rclone('repo')
        mock_call.assert_called_once()

    def test_update_serverwebroot(self):
        """rsync works with file paths, so this test uses paths for the URLs"""
        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir(parents=True)
        fake_apk = repo / 'fake.apk'
        with fake_apk.open('w') as fp:
            fp.write('not an APK, but has the right filename')
        fake_index = repo / fdroidserver.common.INDEX_FILES[0]
        with fake_index.open('w') as fp:
            fp.write('not an index, but has the right filename')
        url = Path('url')
        url.mkdir()

        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.identity_file = None
        fdroidserver.common.options.identity_file = None
        fdroidserver.deploy.config['make_current_version_link'] = False

        dest_apk = Path(url) / fake_apk
        dest_index = Path(url) / fake_index
        self.assertFalse(dest_apk.is_file())
        self.assertFalse(dest_index.is_file())

        fdroidserver.deploy.update_serverwebroot({'url': str(url)}, 'repo')
        self.assertTrue(dest_apk.is_file())
        self.assertTrue(dest_index.is_file())

    def test_update_serverwebroot_in_index_only_mode(self):
        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir()
        fake_apk = repo / 'fake.apk'
        with fake_apk.open('w') as fp:
            fp.write('not an APK, but has the right filename')
        fake_index = repo / fdroidserver.common.INDEX_FILES[0]
        with fake_index.open('w') as fp:
            fp.write('not an index, but has the right filename')
        url = Path('url')
        url.mkdir()

        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.identity_file = None
        fdroidserver.deploy.config['make_current_version_link'] = False

        dest_apk = Path(url) / fake_apk
        dest_index = Path(url) / fake_index
        self.assertFalse(dest_apk.is_file())
        self.assertFalse(dest_index.is_file())

        fdroidserver.deploy.update_serverwebroot(
            {'url': str(url), 'index_only': True}, 'repo'
        )
        self.assertFalse(dest_apk.is_file())
        self.assertTrue(dest_index.is_file())

    @mock.patch.dict(os.environ, clear=True)
    def test_update_serverwebroot_no_rsync_error(self):
        os.environ['PATH'] = self.testdir
        os.chdir(self.testdir)
        with self.assertRaises(fdroidserver.exception.FDroidException):
            fdroidserver.deploy.update_serverwebroot('serverwebroot', 'repo')

    def test_update_serverwebroot_make_cur_version_link(self):
        self.maxDiff = None

        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.no_checksum = True
        fdroidserver.common.options.identity_file = None
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True
        fdroidserver.common.options.index_only = False
        fdroidserver.deploy.config = {'make_current_version_link': True}
        url = "example.com:/var/www/fdroid"
        repo_section = 'repo'

        # setup function for asserting subprocess.call invocations
        call_iteration = 0

        def update_server_webroot_call(cmd):
            nonlocal call_iteration
            if call_iteration == 0:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--quiet',
                        '--exclude',
                        'repo/altstore-index.json',
                        '--exclude',
                        'repo/altstore-index.json.asc',
                        '--exclude',
                        'repo/entry.jar',
                        '--exclude',
                        'repo/entry.json',
                        '--exclude',
                        'repo/entry.json.asc',
                        '--exclude',
                        'repo/index-v1.jar',
                        '--exclude',
                        'repo/index-v1.json',
                        '--exclude',
                        'repo/index-v1.json.asc',
                        '--exclude',
                        'repo/index-v2.json',
                        '--exclude',
                        'repo/index-v2.json.asc',
                        '--exclude',
                        'repo/index.css',
                        '--exclude',
                        'repo/index.html',
                        '--exclude',
                        'repo/index.jar',
                        '--exclude',
                        'repo/index.png',
                        '--exclude',
                        'repo/index.xml',
                        '--exclude',
                        'repo/signer-index.jar',
                        '--exclude',
                        'repo/signer-index.json',
                        '--exclude',
                        'repo/signer-index.json.asc',
                        'repo',
                        'example.com:/var/www/fdroid',
                    ],
                )
            elif call_iteration == 1:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--quiet',
                        'repo',
                        url,
                    ],
                )
            elif call_iteration == 2:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--quiet',
                        'Sym.apk',
                        'Sym.apk.asc',
                        'Sym.apk.sig',
                        'example.com:/var/www/fdroid',
                    ],
                )
            else:
                self.fail('unexpected subprocess.call invocation')
            call_iteration += 1
            return 0

        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            os.mkdir('repo')
            os.symlink('repo/com.example.sym.apk', 'Sym.apk')
            os.symlink('repo/com.example.sym.apk.asc', 'Sym.apk.asc')
            os.symlink('repo/com.example.sym.apk.sig', 'Sym.apk.sig')
            with mock.patch('subprocess.call', side_effect=update_server_webroot_call):
                fdroidserver.deploy.update_serverwebroot({'url': url}, repo_section)
        self.assertEqual(call_iteration, 3, 'expected 3 invocations of subprocess.call')

    def test_update_serverwebroot_make_cur_version_link_in_index_only_mode(self):
        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.no_checksum = True
        fdroidserver.common.options.identity_file = None
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True
        fdroidserver.common.options.identity_file = None
        fdroidserver.deploy.config['make_current_version_link'] = True
        url = "example.com:/var/www/fdroid"
        repo_section = 'repo'

        # setup function for asserting subprocess.call invocations
        call_iteration = 0

        def update_server_webroot_call(cmd):
            nonlocal call_iteration
            if call_iteration == 0:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--quiet',
                        'repo/altstore-index.json',
                        'repo/altstore-index.json.asc',
                        'repo/entry.jar',
                        'repo/entry.json',
                        'repo/entry.json.asc',
                        'repo/index-v1.jar',
                        'repo/index-v1.json',
                        'repo/index-v1.json.asc',
                        'repo/index-v2.json',
                        'repo/index-v2.json.asc',
                        'repo/index.css',
                        'repo/index.html',
                        'repo/index.jar',
                        'repo/index.png',
                        'repo/index.xml',
                        'repo/signer-index.jar',
                        'repo/signer-index.json',
                        'repo/signer-index.json.asc',
                        'example.com:/var/www/fdroid/repo/',
                    ],
                )
            elif call_iteration == 1:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--quiet',
                        'repo',
                        url,
                    ],
                )
            elif call_iteration == 2:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--quiet',
                        'Sym.apk',
                        'Sym.apk.asc',
                        'Sym.apk.sig',
                        'example.com:/var/www/fdroid',
                    ],
                )
            else:
                self.fail('unexpected subprocess.call invocation')
            call_iteration += 1
            return 0

        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            os.mkdir(repo_section)
            os.symlink('repo/com.example.sym.apk', 'Sym.apk')
            os.symlink('repo/com.example.sym.apk.asc', 'Sym.apk.asc')
            os.symlink('repo/com.example.sym.apk.sig', 'Sym.apk.sig')

            fake_files = fdroidserver.common.INDEX_FILES
            for filename in fake_files:
                fake_file = Path(repo_section) / filename
                with fake_file.open('w') as fp:
                    fp.write('not a real one, but has the right filename')

            with mock.patch('subprocess.call', side_effect=update_server_webroot_call):
                fdroidserver.deploy.update_serverwebroot(
                    {'url': url, 'index_only': True}, repo_section
                )
        self.assertEqual(call_iteration, 1, 'expected 1 invocations of subprocess.call')

    def test_update_serverwebroot_with_id_file(self):
        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.identity_file = None
        fdroidserver.common.options.no_checksum = True
        fdroidserver.common.options.verbose = True
        fdroidserver.common.options.quiet = False
        fdroidserver.common.options.identity_file = None
        fdroidserver.common.options.index_only = False
        fdroidserver.deploy.config = {'identity_file': './id_rsa'}
        url = "example.com:/var/www/fdroid"
        repo_section = 'archive'

        # setup function for asserting subprocess.call invocations
        call_iteration = 0

        def update_server_webroot_call(cmd):
            nonlocal call_iteration
            if call_iteration == 0:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--verbose',
                        '-e',
                        'ssh -oBatchMode=yes -oIdentitiesOnly=yes -i '
                        + fdroidserver.deploy.config['identity_file'],
                        '--exclude',
                        'archive/altstore-index.json',
                        '--exclude',
                        'archive/altstore-index.json.asc',
                        '--exclude',
                        'archive/entry.jar',
                        '--exclude',
                        'archive/entry.json',
                        '--exclude',
                        'archive/entry.json.asc',
                        '--exclude',
                        'archive/index-v1.jar',
                        '--exclude',
                        'archive/index-v1.json',
                        '--exclude',
                        'archive/index-v1.json.asc',
                        '--exclude',
                        'archive/index-v2.json',
                        '--exclude',
                        'archive/index-v2.json.asc',
                        '--exclude',
                        'archive/index.css',
                        '--exclude',
                        'archive/index.html',
                        '--exclude',
                        'archive/index.jar',
                        '--exclude',
                        'archive/index.png',
                        '--exclude',
                        'archive/index.xml',
                        'archive',
                        url,
                    ],
                )
            elif call_iteration == 1:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--verbose',
                        '-e',
                        'ssh -oBatchMode=yes -oIdentitiesOnly=yes -i '
                        + fdroidserver.deploy.config['identity_file'],
                        'archive',
                        url,
                    ],
                )
            else:
                self.fail('unexpected subprocess.call invocation')
            call_iteration += 1
            return 0

        with mock.patch('subprocess.call', side_effect=update_server_webroot_call):
            fdroidserver.deploy.update_serverwebroot({'url': url}, repo_section)
        self.assertEqual(call_iteration, 2, 'expected 2 invocations of subprocess.call')

    def test_update_serverwebroot_with_id_file_in_index_only_mode(self):
        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.no_chcksum = False
        fdroidserver.common.options.verbose = True
        fdroidserver.common.options.quiet = False
        fdroidserver.common.options.identity_file = None
        fdroidserver.deploy.config['identity_file'] = './id_rsa'
        fdroidserver.deploy.config['make_current_version_link'] = False
        url = "example.com:/var/www/fdroid"
        repo_section = 'archive'

        # setup function for asserting subprocess.call invocations
        call_iteration = 0

        def update_server_webroot_call(cmd):
            nonlocal call_iteration
            if call_iteration == 0:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--verbose',
                        '-e',
                        'ssh -oBatchMode=yes -oIdentitiesOnly=yes -i '
                        + fdroidserver.deploy.config['identity_file'],
                        'archive/altstore-index.json',
                        'archive/altstore-index.json.asc',
                        'archive/entry.jar',
                        'archive/entry.json',
                        'archive/entry.json.asc',
                        'archive/index-v1.jar',
                        'archive/index-v1.json',
                        'archive/index-v1.json.asc',
                        'archive/index-v2.json',
                        'archive/index-v2.json.asc',
                        'archive/index.css',
                        'archive/index.html',
                        'archive/index.jar',
                        'archive/index.png',
                        'archive/index.xml',
                        "example.com:/var/www/fdroid/archive/",
                    ],
                )
            elif call_iteration == 1:
                self.assertListEqual(
                    cmd,
                    [
                        'rsync',
                        '--archive',
                        '--delete-after',
                        '--safe-links',
                        '--verbose',
                        '-e',
                        'ssh -oBatchMode=yes -oIdentitiesOnly=yes -i '
                        + fdroidserver.deploy.config['identity_file'],
                        "example.com:/var/www/fdroid/archive/",
                    ],
                )
            else:
                self.fail('unexpected subprocess.call invocation')
            call_iteration += 1
            return 0

        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            with mock.patch('subprocess.call', side_effect=update_server_webroot_call):
                os.mkdir(repo_section)
                fake_files = fdroidserver.common.INDEX_FILES
                for filename in fake_files:
                    fake_file = Path(repo_section) / filename
                    with fake_file.open('w') as fp:
                        fp.write('not a real one, but has the right filename')

                fdroidserver.deploy.update_serverwebroot(
                    {'url': url, 'index_only': True}, repo_section
                )
        self.assertEqual(call_iteration, 1, 'expected 1 invocations of subprocess.call')

    @unittest.skipIf(
        not os.getenv('VIRUSTOTAL_API_KEY'), 'VIRUSTOTAL_API_KEY is not set'
    )
    def test_upload_to_virustotal(self):
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.verbose = True
        virustotal_apikey = os.getenv('VIRUSTOTAL_API_KEY')
        fdroidserver.deploy.upload_to_virustotal('repo', virustotal_apikey)

    def test_remote_hostname_regex(self):
        for remote_url, name in (
            ('git@github.com:guardianproject/fdroid-repo', 'github'),
            ('git@gitlab.com:guardianproject/fdroid-repo', 'gitlab'),
            ('https://github.com:guardianproject/fdroid-repo', 'github'),
            ('https://gitlab.com/guardianproject/fdroid-repo', 'gitlab'),
            ('https://salsa.debian.org/foo/repo', 'salsa'),
        ):
            self.assertEqual(
                name, fdroidserver.deploy.REMOTE_HOSTNAME_REGEX.sub(r'\1', remote_url)
            )

    def test_update_awsbucket_s3cmd(self):
        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.no_checksum = True
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True

        config = {}
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.deploy.config = config
        fdroidserver.deploy.config["awsbucket"] = "bucket"
        fdroidserver.deploy.config["awsaccesskeyid"] = "accesskeyid"
        fdroidserver.deploy.config["awssecretkey"] = "secretkey"
        fdroidserver.deploy.config["s3cmd"] = "s3cmd"

        repo_section = 'repo'

        # setup function for asserting subprocess.call invocations
        call_iteration = 0

        def update_awsbucket_s3cmd_call(cmd):
            nonlocal call_iteration
            if call_iteration == 0:
                self.assertListEqual(
                    cmd,
                    [
                        's3cmd',
                        f"--config={fdroidserver.deploy.AUTO_S3CFG}",
                        'info',
                        f"s3://{fdroidserver.deploy.config['awsbucket']}",
                    ],
                )
            elif call_iteration == 1:
                self.assertListEqual(
                    cmd,
                    [
                        's3cmd',
                        f"--config={fdroidserver.deploy.AUTO_S3CFG}",
                        'sync',
                        '--acl-public',
                        '--quiet',
                        '--exclude',
                        'repo/altstore-index.json',
                        '--exclude',
                        'repo/altstore-index.json.asc',
                        '--exclude',
                        'repo/entry.jar',
                        '--exclude',
                        'repo/entry.json',
                        '--exclude',
                        'repo/entry.json.asc',
                        '--exclude',
                        'repo/index-v1.jar',
                        '--exclude',
                        'repo/index-v1.json',
                        '--exclude',
                        'repo/index-v1.json.asc',
                        '--exclude',
                        'repo/index-v2.json',
                        '--exclude',
                        'repo/index-v2.json.asc',
                        '--exclude',
                        'repo/index.css',
                        '--exclude',
                        'repo/index.html',
                        '--exclude',
                        'repo/index.jar',
                        '--exclude',
                        'repo/index.png',
                        '--exclude',
                        'repo/index.xml',
                        '--exclude',
                        'repo/signer-index.jar',
                        '--exclude',
                        'repo/signer-index.json',
                        '--exclude',
                        'repo/signer-index.json.asc',
                        '--no-check-md5',
                        '--skip-existing',
                        repo_section,
                        f"s3://{fdroidserver.deploy.config['awsbucket']}/fdroid/",
                    ],
                )
            elif call_iteration == 2:
                self.assertListEqual(
                    cmd,
                    [
                        's3cmd',
                        f"--config={fdroidserver.deploy.AUTO_S3CFG}",
                        'sync',
                        '--acl-public',
                        '--quiet',
                        '--exclude',
                        'repo/altstore-index.json',
                        '--exclude',
                        'repo/altstore-index.json.asc',
                        '--exclude',
                        'repo/entry.jar',
                        '--exclude',
                        'repo/entry.json',
                        '--exclude',
                        'repo/entry.json.asc',
                        '--exclude',
                        'repo/index-v1.jar',
                        '--exclude',
                        'repo/index-v1.json',
                        '--exclude',
                        'repo/index-v1.json.asc',
                        '--exclude',
                        'repo/index-v2.json',
                        '--exclude',
                        'repo/index-v2.json.asc',
                        '--exclude',
                        'repo/index.css',
                        '--exclude',
                        'repo/index.html',
                        '--exclude',
                        'repo/index.jar',
                        '--exclude',
                        'repo/index.png',
                        '--exclude',
                        'repo/index.xml',
                        '--exclude',
                        'repo/signer-index.jar',
                        '--exclude',
                        'repo/signer-index.json',
                        '--exclude',
                        'repo/signer-index.json.asc',
                        '--no-check-md5',
                        repo_section,
                        f"s3://{fdroidserver.deploy.config['awsbucket']}/fdroid/",
                    ],
                )
            elif call_iteration == 3:
                self.assertListEqual(
                    cmd,
                    [
                        's3cmd',
                        f"--config={fdroidserver.deploy.AUTO_S3CFG}",
                        'sync',
                        '--acl-public',
                        '--quiet',
                        '--delete-removed',
                        '--delete-after',
                        '--no-check-md5',
                        repo_section,
                        f"s3://{fdroidserver.deploy.config['awsbucket']}/fdroid/",
                    ],
                )
            else:
                self.fail('unexpected subprocess.call invocation')
            call_iteration += 1
            return 0

        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            os.mkdir('repo')
            os.symlink('repo/com.example.sym.apk', 'Sym.apk')
            os.symlink('repo/com.example.sym.apk.asc', 'Sym.apk.asc')
            os.symlink('repo/com.example.sym.apk.sig', 'Sym.apk.sig')
            with mock.patch('subprocess.call', side_effect=update_awsbucket_s3cmd_call):
                fdroidserver.deploy.update_awsbucket_s3cmd(repo_section)
        self.assertEqual(call_iteration, 4, 'expected 4 invocations of subprocess.call')

    def test_update_awsbucket_s3cmd_in_index_only_mode(self):
        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.no_checksum = True
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True

        config = {}
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.deploy.config = config
        fdroidserver.deploy.config["awsbucket"] = "bucket"
        fdroidserver.deploy.config["awsaccesskeyid"] = "accesskeyid"
        fdroidserver.deploy.config["awssecretkey"] = "secretkey"
        fdroidserver.deploy.config["s3cmd"] = "s3cmd"

        repo_section = 'repo'

        # setup function for asserting subprocess.call invocations
        call_iteration = 0

        def update_awsbucket_s3cmd_call(cmd):
            nonlocal call_iteration
            if call_iteration == 0:
                self.assertListEqual(
                    cmd,
                    [
                        's3cmd',
                        f"--config={fdroidserver.deploy.AUTO_S3CFG}",
                        'info',
                        f"s3://{fdroidserver.deploy.config['awsbucket']}",
                    ],
                )
            elif call_iteration == 1:
                self.assertListEqual(
                    cmd,
                    [
                        's3cmd',
                        f"--config={fdroidserver.deploy.AUTO_S3CFG}",
                        'sync',
                        '--acl-public',
                        '--quiet',
                        '--include',
                        'repo/altstore-index.json',
                        '--include',
                        'repo/altstore-index.json.asc',
                        '--include',
                        'repo/entry.jar',
                        '--include',
                        'repo/entry.json',
                        '--include',
                        'repo/entry.json.asc',
                        '--include',
                        'repo/index-v1.jar',
                        '--include',
                        'repo/index-v1.json',
                        '--include',
                        'repo/index-v1.json.asc',
                        '--include',
                        'repo/index-v2.json',
                        '--include',
                        'repo/index-v2.json.asc',
                        '--include',
                        'repo/index.css',
                        '--include',
                        'repo/index.html',
                        '--include',
                        'repo/index.jar',
                        '--include',
                        'repo/index.png',
                        '--include',
                        'repo/index.xml',
                        '--include',
                        'repo/signer-index.jar',
                        '--include',
                        'repo/signer-index.json',
                        '--include',
                        'repo/signer-index.json.asc',
                        '--delete-removed',
                        '--delete-after',
                        '--no-check-md5',
                        repo_section,
                        f"s3://{fdroidserver.deploy.config['awsbucket']}/fdroid/",
                    ],
                )
            else:
                self.fail('unexpected subprocess.call invocation')
            call_iteration += 1
            return 0

        with tempfile.TemporaryDirectory() as tmpdir, TmpCwd(tmpdir):
            os.mkdir('repo')
            os.symlink('repo/com.example.sym.apk', 'Sym.apk')
            os.symlink('repo/com.example.sym.apk.asc', 'Sym.apk.asc')
            os.symlink('repo/com.example.sym.apk.sig', 'Sym.apk.sig')
            with mock.patch('subprocess.call', side_effect=update_awsbucket_s3cmd_call):
                fdroidserver.deploy.update_awsbucket_s3cmd(
                    repo_section, is_index_only=True
                )
        self.assertEqual(call_iteration, 2, 'expected 2 invocations of subprocess.call')

    def test_update_awsbucket_libcloud(self):
        from libcloud.storage.base import Container

        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.no_checksum = True
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True

        config = {}
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.deploy.config = config
        fdroidserver.deploy.config["awsbucket"] = "bucket"
        fdroidserver.deploy.config["awsaccesskeyid"] = "accesskeyid"
        fdroidserver.deploy.config["awssecretkey"] = "secretkey"
        fdroidserver.deploy.config["s3cmd"] = "s3cmd"

        repo_section = 'repo'

        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir(parents=True)
        fake_apk = repo / 'Sym.apk'
        with fake_apk.open('w') as fp:
            fp.write('not an APK, but has the right filename')
        fake_index = repo / fdroidserver.common.INDEX_FILES[0]
        with fake_index.open('w') as fp:
            fp.write('not an index, but has the right filename')

        with mock.patch(
            'libcloud.storage.drivers.s3.S3StorageDriver'
        ) as mock_driver_class:
            mock_driver = mock_driver_class.return_value
            mock_container = mock.MagicMock(spec=Container)
            mock_container.list_objects.return_value = [
                mock.MagicMock(name='Sym.apk'),
                mock.MagicMock(name=fdroidserver.common.INDEX_FILES[0]),
            ]

            mock_driver.get_container.return_value = mock_container
            mock_driver.upload_object_via_stream.return_value = None

            fdroidserver.deploy.update_awsbucket_libcloud(repo_section)

            mock_driver.get_container.assert_called_once_with(
                container_name=fdroidserver.deploy.config["awsbucket"]
            )
            mock_container.list_objects.assert_called_once_with()
            files_to_upload = [
                'fdroid/repo/Sym.apk',
                f"fdroid/repo/{fdroidserver.common.INDEX_FILES[0]}",
            ]
            calls = [
                mock.call(
                    iterator=mock.ANY,
                    container=mock_container,
                    object_name=file,
                    extra={'acl': 'public-read'},
                )
                for file in files_to_upload
            ]
            mock_driver.upload_object_via_stream.assert_has_calls(calls, any_order=True)
            self.assertEqual(mock_driver.upload_object_via_stream.call_count, 2)

    def test_update_awsbucket_libcloud_in_index_only_mode(self):
        from libcloud.storage.base import Container

        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.no_checksum = True
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True

        config = {}
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.deploy.config = config
        fdroidserver.deploy.config["awsbucket"] = "bucket"
        fdroidserver.deploy.config["awsaccesskeyid"] = "accesskeyid"
        fdroidserver.deploy.config["awssecretkey"] = "secretkey"
        fdroidserver.deploy.config["s3cmd"] = "s3cmd"

        repo_section = 'repo'

        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir(parents=True)
        fake_apk = repo / 'Sym.apk'
        with fake_apk.open('w') as fp:
            fp.write('not an APK, but has the right filename')
        fake_index = repo / fdroidserver.common.INDEX_FILES[0]
        with fake_index.open('w') as fp:
            fp.write('not an index, but has the right filename')

        with mock.patch(
            'libcloud.storage.drivers.s3.S3StorageDriver'
        ) as mock_driver_class:
            mock_driver = mock_driver_class.return_value
            mock_container = mock.MagicMock(spec=Container)
            mock_container.list_objects.return_value = [
                mock.MagicMock(name='Sym.apk'),
                mock.MagicMock(name=fdroidserver.common.INDEX_FILES[0]),
            ]

            mock_driver.get_container.return_value = mock_container
            mock_driver.upload_object_via_stream.return_value = None

            fdroidserver.deploy.update_awsbucket_libcloud(
                repo_section, is_index_only=True
            )

            mock_driver.get_container.assert_called_once_with(
                container_name=fdroidserver.deploy.config["awsbucket"]
            )
            mock_container.list_objects.assert_called_once_with()
            files_to_upload = [f"fdroid/repo/{fdroidserver.common.INDEX_FILES[0]}"]
            calls = [
                mock.call(
                    iterator=mock.ANY,
                    container=mock_container,
                    object_name=file,
                    extra={'acl': 'public-read'},
                )
                for file in files_to_upload
            ]
            mock_driver.upload_object_via_stream.assert_has_calls(
                calls,
                any_order=False,
            )
            self.assertEqual(mock_driver.upload_object_via_stream.call_count, 1)

    def test_update_servergitmirrors(self):
        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.identity_file = None
        fdroidserver.common.options.no_keep_git_mirror_archive = False
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True

        config = {}
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.deploy.config = config

        os.chdir(self.testdir)

        repo_section = 'repo'
        initial_branch = fdroidserver.deploy.GIT_BRANCH

        remote_repo = Path(self.testdir) / 'remote'
        remote_repo.mkdir(parents=True)
        remote_git_repo = git.Repo.init(
            remote_repo, initial_branch=initial_branch, bare=True
        )
        fdroidserver.deploy.config["servergitmirrors"] = [{"url": str(remote_repo)}]

        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir(parents=True)
        fake_apk = 'Sym.apk'
        fake_files = fdroidserver.common.INDEX_FILES + [fake_apk]
        for filename in fake_files:
            fake_file = repo / filename
            with fake_file.open('w') as fp:
                fp.write('not a real one, but has the right filename')

        fdroidserver.deploy.update_servergitmirrors(
            fdroidserver.deploy.config["servergitmirrors"], repo_section
        )

        verify_repo = remote_git_repo.clone(
            Path(self.testdir) / 'verify',
        )

        for filename in fake_files:
            remote_file = f"fdroid/{repo_section}/{filename}"

            self.assertIsNotNone(verify_repo.working_tree_dir)
            if verify_repo.working_tree_dir is not None:
                self.assertTrue(
                    (Path(verify_repo.working_tree_dir) / remote_file).exists()
                )

    def test_update_servergitmirrors_in_index_only_mode(self):
        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.identity_file = None
        fdroidserver.common.options.no_keep_git_mirror_archive = False
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True

        config = {}
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.deploy.config = config

        os.chdir(self.testdir)

        repo_section = 'repo'
        initial_branch = fdroidserver.deploy.GIT_BRANCH

        remote_repo = Path(self.testdir) / 'remote'
        remote_repo.mkdir(parents=True)
        remote_git_repo = git.Repo.init(
            remote_repo, initial_branch=initial_branch, bare=True
        )
        fdroidserver.deploy.config["servergitmirrors"] = [
            {"url": str(remote_repo), "index_only": True}
        ]

        os.chdir(self.testdir)
        repo = Path('repo')
        repo.mkdir(parents=True)
        fake_apk = 'Sym.apk'
        fake_files = fdroidserver.common.INDEX_FILES + [fake_apk]
        for filename in fake_files:
            fake_file = repo / filename
            with fake_file.open('w') as fp:
                fp.write('not a real one, but has the right filename')

        fdroidserver.deploy.update_servergitmirrors(
            fdroidserver.deploy.config["servergitmirrors"], repo_section
        )

        verify_repo = remote_git_repo.clone(
            Path(self.testdir) / 'verify',
        )

        for filename in fdroidserver.common.INDEX_FILES:
            remote_file = f"fdroid/{repo_section}/{filename}"

            self.assertIsNotNone(verify_repo.working_tree_dir)
            if verify_repo.working_tree_dir is not None:
                self.assertTrue(
                    (Path(verify_repo.working_tree_dir) / remote_file).exists()
                )

        # Should not have the APK file
        remote_file = f"fdroid/{repo_section}/{fake_apk}"
        if verify_repo.working_tree_dir is not None:
            self.assertFalse(
                (Path(verify_repo.working_tree_dir) / remote_file).exists()
            )

    def test_upload_to_servergitmirror_in_index_only_mode(self):
        # setup parameters for this test run
        fdroidserver.common.options = mock.Mock()
        fdroidserver.common.options.identity_file = None
        fdroidserver.common.options.no_keep_git_mirror_archive = False
        fdroidserver.common.options.verbose = False
        fdroidserver.common.options.quiet = True
        fdroidserver.common.options.identity_file = None

        config = {}
        fdroidserver.common.fill_config_defaults(config)
        fdroidserver.deploy.config = config

        repo_section = 'repo'
        initial_branch = fdroidserver.deploy.GIT_BRANCH

        os.chdir(self.testdir)

        local_git_repo_path = Path(self.testdir) / 'local'
        local_git_repo = git.Repo.init(
            local_git_repo_path, initial_branch=initial_branch
        )

        fdroid_dir = local_git_repo_path / 'fdroid'
        repo_dir = fdroid_dir / repo_section
        repo_dir.mkdir(parents=True)
        fake_apk = 'Sym.apk'
        fake_files = fdroidserver.common.INDEX_FILES + [fake_apk]
        for filename in fake_files:
            fake_file = repo_dir / filename
            with fake_file.open('w') as fp:
                fp.write('not a real one, but has the right filename')

        # The remote repo must be a bare repo to allow being pushed to
        remote_git_repo_dir = Path(self.testdir) / 'remote'
        remote_git_repo = git.Repo.init(
            remote_git_repo_dir, initial_branch=initial_branch, bare=True
        )

        mirror_config = {"url": str(remote_git_repo_dir), "index_only": True}
        enabled_remotes = []
        ssh_cmd = 'ssh -oBatchMode=yes'
        fdroidserver.deploy.upload_to_servergitmirror(
            mirror_config=mirror_config,
            local_repo=local_git_repo,
            enabled_remotes=enabled_remotes,
            repo_section=repo_section,
            is_index_only=mirror_config['index_only'],
            fdroid_dir=str(fdroid_dir),
            git_mirror_path=str(local_git_repo_path),
            ssh_cmd=ssh_cmd,
            progress=git.RemoteProgress(),
        )

        verify_repo = remote_git_repo.clone(
            Path(self.testdir) / 'verify',
        )

        for filename in fdroidserver.common.INDEX_FILES:
            remote_file = f"fdroid/{repo_section}/{filename}"

            self.assertIsNotNone(verify_repo.working_tree_dir)
            if verify_repo.working_tree_dir is not None:
                self.assertTrue(
                    (Path(verify_repo.working_tree_dir) / remote_file).exists()
                )

        # Should not have the APK file
        remote_file = f"fdroid/{repo_section}/{fake_apk}"
        if verify_repo.working_tree_dir is not None:
            self.assertFalse(
                (Path(verify_repo.working_tree_dir) / remote_file).exists()
            )


class GitHubReleasesTest(unittest.TestCase):
    def test_find_release_infos(self):
        self.maxDiff = None

        index_mock = b"""
            {
                "packages": {
                    "com.example.app": {
                        "versions": {
                            "2e6f263c1927506015bfc98bce0818247836f2e7fe29a04e1af2b33c97848750": {
                                "file": {
                                    "name": "/com.example.app_123.apk"
                                },
                                "whatsNew": {
                                    "en-US": "fake what's new"
                                },
                                "manifest": {
                                    "versionName": "1.2.3",
                                    "versionCode": "123"
                                }
                            },
                            "8a6f263c8327506015bfc98bce0815247836f2e7fe29a04e1af2bffa6409998d": {
                                "file": {
                                    "name": "/com.example.app_100.apk"
                                },
                                "manifest": {
                                    "versionName": "1.0-alpha",
                                    "versionCode": "123"
                                },
                                "releaseChannels": ["alpha"]
                            }
                        }
                    },
                    "another.app": {
                        "versions": {
                            "30602ffc19a7c0601bbfa93bce00082c78a6f2ddfe29a04e1af253fc9f84eda0": {
                                "file": {
                                    "name": "/another.app_1.apk"
                                },
                                "manifest": {
                                    "versionName": "1",
                                    "versionCode": "1"
                                }
                            }
                        }
                    },
                    "fildered.app": {
                        "versions": {
                            "93ae02fc19a7c0601adfa93bce0443fc78a6f2ddfe3df04e1af093fca9a1ff09": {
                                "file": {
                                    "name": "/another.app_1.apk"
                                },
                                "manifest": {
                                    "versionName": "1",
                                    "versionCode": "1"
                                }
                            }
                        }
                    }
                }
            }
        """
        with unittest.mock.patch(
            "fdroidserver.deploy.open", unittest.mock.mock_open(read_data=index_mock)
        ):
            release_infos = fdroidserver.deploy.find_release_infos(
                "fake_path",
                Path('fake_repo'),
                ["com.example.app", "another.app"],
            )

        self.assertDictEqual(
            release_infos,
            {
                "another.app": {
                    "1": {
                        "files": [Path('fake_repo') / "another.app_1.apk"],
                        "hasReleaseChannels": False,
                        "whatsNew": None,
                    },
                },
                "com.example.app": {
                    "1.0-alpha": {
                        "files": [
                            Path("fake_repo") / "com.example.app_100.apk",
                        ],
                        "hasReleaseChannels": True,
                        "whatsNew": None,
                    },
                    "1.2.3": {
                        "files": [
                            Path("fake_repo") / "com.example.app_123.apk",
                        ],
                        "hasReleaseChannels": False,
                        "whatsNew": "fake what's new",
                    },
                },
            },
        )

    def test_upload_to_github_releases(self):
        gh_config = [
            {
                "projectUrl": "https://github.com/example/app",
                "packageNames": ["com.example.app", "another.app"],
            },
            {
                "projectUrl": "https://github.com/custom/app",
                "packageNames": ["more.custom.app"],
                "token": "custom_token",
            },
        ]

        fri_mock = unittest.mock.Mock(return_value="fri_result")
        urr_mock = unittest.mock.Mock()
        with unittest.mock.patch(
            "fdroidserver.deploy.find_release_infos", fri_mock
        ), unittest.mock.patch(
            "fdroidserver.deploy.upload_to_github_releases_repo", urr_mock
        ), tempfile.TemporaryDirectory() as tmpdir:
            with open(Path(tmpdir) / "index-v2.json", "w") as f:
                f.write("")

            fdroidserver.deploy.upload_to_github_releases(
                tmpdir, gh_config, "fake_global_token"
            )

            fri_mock.assert_called_once_with(
                Path(tmpdir) / "index-v2.json",
                Path(tmpdir),
                ["com.example.app", "another.app", "more.custom.app"],
            )

        self.maxDiff = None
        self.assertListEqual(
            urr_mock.call_args_list,
            [
                unittest.mock.call(
                    {
                        "projectUrl": "https://github.com/example/app",
                        "packageNames": ["com.example.app", "another.app"],
                    },
                    "fri_result",
                    "fake_global_token",
                ),
                unittest.mock.call(
                    {
                        "projectUrl": "https://github.com/custom/app",
                        "packageNames": ["more.custom.app"],
                        "token": "custom_token",
                    },
                    "fri_result",
                    "fake_global_token",
                ),
            ],
        )


class Test_UploadToGithubReleasesRepo(unittest.TestCase):
    def setUp(self):
        self.repo_conf = {
            "projectUrl": "https://github.com/example/app",
            "packageNames": ["com.example.app", "com.example.altapp", "another.app"],
        }
        self.release_infos = {
            "com.example.app": {
                "1.0.0": {
                    "files": [
                        Path("fake_repo") / "com.example.app_100100.apk",
                    ],
                    "hasReleaseChannels": False,
                    "whatsNew": "what's new com.example.app 1.0.0",
                },
                "1.0.0-beta1": {
                    "files": [
                        Path("fake_repo") / "com.example.app_100007.apk",
                    ],
                    "hasReleaseChannels": True,
                    "whatsNew": None,
                },
            },
            "com.example.altapp": {
                "1.0.0": {
                    "files": [
                        Path("fake_repo") / "com.example.altapp_100100.apk",
                        Path("fake_repo") / "com.example.altapp_100100.apk.asc",
                        Path("fake_repo") / "com.example.altapp_100100.apk.idsig",
                    ],
                    "whatsNew": "what's new com.example.altapp 1.0.0",
                },
            },
        }

        self.api = unittest.mock.Mock()
        self.api.list_unreleased_tags = lambda: ["1.0.0", "1.0.0-beta1"]
        self.api_constructor = unittest.mock.Mock(return_value=self.api)

    def test_global_token(self):
        with unittest.mock.patch("fdroidserver.github.GithubApi", self.api_constructor):
            fdroidserver.deploy.upload_to_github_releases_repo(
                self.repo_conf,
                self.release_infos,
                "global_token",
            )

        self.api_constructor.assert_called_once_with(
            "global_token", "https://github.com/example/app"
        )

        self.assertListEqual(
            self.api.create_release.call_args_list,
            [
                unittest.mock.call(
                    "1.0.0",
                    [
                        Path("fake_repo/com.example.app_100100.apk"),
                        Path("fake_repo/com.example.altapp_100100.apk"),
                        Path("fake_repo/com.example.altapp_100100.apk.asc"),
                        Path("fake_repo/com.example.altapp_100100.apk.idsig"),
                    ],
                    "what's new com.example.app 1.0.0",
                ),
            ],
        )

    def test_local_token(self):
        self.repo_conf["token"] = "local_token"  # nosec B105
        with unittest.mock.patch("fdroidserver.github.GithubApi", self.api_constructor):
            fdroidserver.deploy.upload_to_github_releases_repo(
                self.repo_conf,
                self.release_infos,
                "global_token",
            )

        self.api_constructor.assert_called_once_with(
            "local_token", "https://github.com/example/app"
        )

        self.assertListEqual(
            self.api.create_release.call_args_list,
            [
                unittest.mock.call(
                    "1.0.0",
                    [
                        Path("fake_repo/com.example.app_100100.apk"),
                        Path("fake_repo/com.example.altapp_100100.apk"),
                        Path("fake_repo/com.example.altapp_100100.apk.asc"),
                        Path("fake_repo/com.example.altapp_100100.apk.idsig"),
                    ],
                    "what's new com.example.app 1.0.0",
                ),
            ],
        )



import gettext
import glob
import os
import sys


# support running straight from git and standard installs
rootpaths = [
    os.path.realpath(os.path.join(os.path.dirname(__file__), '..')),
    os.path.realpath(
        os.path.join(os.path.dirname(__file__), '..', '..', '..', '..', 'share')
    ),
    os.path.join(sys.prefix, 'share'),
]

localedir = None
for rootpath in rootpaths:
    if len(glob.glob(os.path.join(rootpath, 'locale', '*', 'LC_MESSAGES', 'fdroidserver.mo'))) > 0:
        localedir = os.path.join(rootpath, 'locale')
        break

gettext.bindtextdomain('fdroidserver', localedir)
gettext.textdomain('fdroidserver')
_ = gettext.gettext


from fdroidserver.exception import (FDroidException,
                                    MetaDataException,
                                    VerificationException)  # NOQA: E402
FDroidException  # NOQA: B101
MetaDataException  # NOQA: B101
VerificationException  # NOQA: B101

from fdroidserver.common import (verify_apk_signature,
                                 genkeystore as generate_keystore)  # NOQA: E402
verify_apk_signature  # NOQA: B101
generate_keystore  # NOQA: B101
from fdroidserver.index import (download_repo_index,
                                download_repo_index_v1,
                                download_repo_index_v2,
                                get_mirror_service_urls,
                                make as make_index)  # NOQA: E402
download_repo_index  # NOQA: B101
download_repo_index_v1  # NOQA: B101
download_repo_index_v2  # NOQA: B101
get_mirror_service_urls  # NOQA: B101
make_index  # NOQA: B101
from fdroidserver.update import (process_apk,
                                 process_apks,
                                 scan_apk,
                                 scan_repo_files)  # NOQA: E402
process_apk  # NOQA: B101
process_apks  # NOQA: B101
scan_apk  # NOQA: B101
scan_repo_files  # NOQA: B101
from fdroidserver.deploy import (update_awsbucket,
                                 update_servergitmirrors,
                                 update_serverwebroots,
                                 update_serverwebroot)  # NOQA: E402
update_awsbucket  # NOQA: B101
update_servergitmirrors  # NOQA: B101
update_serverwebroots  # NOQA: B101
update_serverwebroot  # NOQA: B101
